/*-
 * Copyright 2021 Hochschule für Technik Stuttgart
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.hft.stuttgart.citygml.viewer;

import java.nio.FloatBuffer;

import org.joml.Math;
import org.joml.Matrix4f;
import org.joml.Vector3f;
import org.lwjgl.system.MemoryUtil;

public class Camera {

	private static final float MIN_DISTANCE = 1f;
	private static final float PI = (float) Math.PI;

	private Matrix4f projMatrix;
	private Matrix4f viewMatrix;
	private Matrix4f projViewMatrix;

	private float rotateAroundX = 2f;
	private float rotateAroundZ = 3f;

	private Vector3f eyeVec = new Vector3f();
	private Vector3f center = new Vector3f();
	private Vector3f up = new Vector3f();
	private float distance = 5;

	private static final float SCROLL_PERCENTAGE = 0.1f;

	private UniformLocation uniformLocation;
	private FloatBuffer matrixBuffer = MemoryUtil.memAllocFloat(16);

	private boolean isOrtho = false;

	public Camera(Shader shader) {
		uniformLocation = shader.getUniformLocation("projViewModel");
		projViewMatrix = new Matrix4f();
		viewMatrix = new Matrix4f();
		projMatrix = new Matrix4f();
		up.z = 1;
	}

	public void reshape(int width, int height, float distance) {
		if (isOrtho) {
			projMatrix.setOrtho(0, width, height, 0, 0, distance);
		} else {
			// 90 degree fow
			float fow = PI / 2f;
			float aspectRatio = (float) width / height;
			projMatrix.setPerspective(fow, aspectRatio, MIN_DISTANCE, distance);

		}
		updateMatrix();
	}

	public void setOrtho(boolean isOrtho) {
		this.isOrtho = isOrtho;
	}

	private void updateMatrix() {
		// view matrix
		if (isOrtho) {
			projMatrix.get(matrixBuffer);
		} else {
			float sx = Math.sin(rotateAroundX);
			float cx = Math.cos(rotateAroundX);
			float sy = Math.sin(rotateAroundZ);
			float cy = Math.cos(rotateAroundZ);
			float sxDistance = distance * sx;
			eyeVec.x = sxDistance * cy;
			eyeVec.y = sxDistance * sy;
			eyeVec.z = distance * cx;
			viewMatrix.identity();
			viewMatrix.translate(0, 0, distance);
			viewMatrix.lookAt(eyeVec, center, up);
			projMatrix.mul(viewMatrix, projViewMatrix);
			// model matrix is identity
			projViewMatrix.get(matrixBuffer);

		}
		uniformLocation.uploadMat4(matrixBuffer);
	}

	public void drag(double dragDiffX, double dragDiffY) {
		rotateAroundX += dragDiffY / 500f;
		rotateAroundZ -= dragDiffX / 500f;
		if (rotateAroundX > Math.PI) {
			rotateAroundX = (float) Math.PI - 0.001f;
		} else if (rotateAroundX < 0.001) {
			rotateAroundX = (float) 0.001;
		}
		updateMatrix();
	}

	public void setDistance(float distance) {
		this.distance = distance;
		updateMatrix();
	}

	public void scroll(int scroll) {
		int addedDistance = (int) (distance * scroll * SCROLL_PERCENTAGE);
		distance += addedDistance;
		updateMatrix();
	}

	public void free() {
		MemoryUtil.memFree(matrixBuffer);
	}

	public void move(double dragDiffX, double dragDiffY) {
	}

}
