/*-
 * Copyright 2021 Hochschule für Technik Stuttgart
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.hft.stuttgart.citygml.viewer;

import static org.lwjgl.opengl.GL11.GL_FLOAT;

import java.awt.Color;
import java.nio.Buffer;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.ArrayList;
import java.util.List;

import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL15;
import org.lwjgl.opengl.GL20;
import org.lwjgl.opengl.GL30;
import org.lwjgl.system.MemoryUtil;

import de.hft.stuttgart.citygml.viewer.datastructure.Polygon;
import de.hft.stuttgart.citygml.viewer.datastructure.TesselatedPolygon;
import de.hft.stuttgart.citygml.viewer.datastructure.Triangle3d;
import de.hft.stuttgart.citygml.viewer.math.Vector3d;

public class PolygonViewInformation {
	
	private static final Vector3d AXIS = new Vector3d(19, 0.8, 1.5).normalize();

	private int vao;
	private boolean draw = false;
	private int elements;
	
	private PolygonListener listener;

	private int posVbo;
	private int colorVbo;
	private int indexVbo;

	public PolygonViewInformation(List<Polygon> polygons, PolygonListener listener) {
		this.listener = listener;
		vao = GL30.glGenVertexArrays();
		GL30.glBindVertexArray(vao);
		
		posVbo = GL15.glGenBuffers();
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, posVbo);
		GL20.glEnableVertexAttribArray(0);
		GL20.glVertexAttribPointer(0, 3, GL_FLOAT, false, 0, 0);
		
		colorVbo = GL15.glGenBuffers();
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, colorVbo);
		GL20.glEnableVertexAttribArray(1);
		GL20.glVertexAttribPointer(1, 3, GL_FLOAT, false, 0, 0);
		
		indexVbo = GL15.glGenBuffers();
		setupBuffers(polygons, posVbo, colorVbo, indexVbo);
		GL30.glBindVertexArray(0);
	}
	
	private void setupBuffers(List<Polygon> polys, int posVbo, int colorVbo, int indexVbo) {
		List<TesselatedPolygon> tessPolys = new ArrayList<>();
		for (Polygon poly : polys) {
			tessPolys.add(TesselatedPolygon.of(poly));
			listener.finishedPolygon();
		}
		
		int nrOfTriangles = 0;
		for (TesselatedPolygon tessPoly : tessPolys) {
			nrOfTriangles += tessPoly.getTriangles().size();
		}

		// 6 float for 3 vertices per triangle
		int vertexDataSize = nrOfTriangles * 3 * 3;
		FloatBuffer vertexData = MemoryUtil.memAllocFloat(vertexDataSize);
		int colorDataSize = nrOfTriangles * 3 * 3;
		FloatBuffer colorData = MemoryUtil.memAllocFloat(colorDataSize);
		
		IntBuffer elementData = MemoryUtil.memAllocInt(nrOfTriangles * 3);
		float[] rgba = new float[4];
		for (TesselatedPolygon tessPoly : tessPolys) {
			Color baseColor = tessPoly.getColor();
			double cos = tessPoly.getNormal().dot(AXIS);
			double acos = Math.acos(cos);
			// normalize to range [0.3, 0.9]
			acos = acos / Math.PI;
			acos = acos * 0.6 + 0.3;
			Color derivedColor = new Color((int) (baseColor.getRed() * acos), (int) (baseColor.getGreen() * acos),
					(int) (baseColor.getBlue() * acos));
			derivedColor.getComponents(rgba);
			for (Triangle3d tri : tessPoly.getTriangles()) {
				addPoint(vertexData, colorData, tri.getP1(), rgba[0], rgba[1], rgba[2]);
				addPoint(vertexData, colorData, tri.getP2(), rgba[0], rgba[1], rgba[2]);
				addPoint(vertexData, colorData, tri.getP3(), rgba[0], rgba[1], rgba[2]);
			}
			listener.finishedPolygon();
		}
		for (int i = 0; i < elementData.capacity(); i++) {
			elementData.put(i);
		}
		((Buffer) elementData).flip();
		((Buffer) colorData).flip();
		((Buffer) vertexData).flip();

		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, posVbo);
		GL15.glBufferData(GL15.GL_ARRAY_BUFFER, vertexData, GL15.GL_STATIC_DRAW);
		
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, colorVbo);
		GL15.glBufferData(GL15.GL_ARRAY_BUFFER, colorData, GL15.GL_STATIC_DRAW);
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, 0);

		GL15.glBindBuffer(GL15.GL_ELEMENT_ARRAY_BUFFER, indexVbo);
		GL15.glBufferData(GL15.GL_ELEMENT_ARRAY_BUFFER, elementData, GL15.GL_STATIC_DRAW);
		GL15.glBindBuffer(GL15.GL_ELEMENT_ARRAY_BUFFER, 0);
	
		MemoryUtil.memFree(elementData);
		MemoryUtil.memFree(colorData);
		MemoryUtil.memFree(vertexData);

		draw = true;
		elements = elementData.capacity();
	}
	
	private static void addPoint(FloatBuffer vertexData, FloatBuffer colorData, Vector3d point, float red, float green, float blue) {
		vertexData.put((float) point.getX());
		vertexData.put((float) point.getY());
		vertexData.put((float) point.getZ());
		colorData.put(red);
		colorData.put(green);
		colorData.put(blue);
	}
	
	public boolean shouldDraw() {
		return draw;
	}
	
	public void setShouldDraw(boolean draw) {
		this.draw = draw;
	}
	
	public int getVao() {
		return vao;
	}
	
	public void destroy() {
		GL15.glDeleteBuffers(colorVbo);
		GL15.glDeleteBuffers(posVbo);
		GL15.glDeleteBuffers(indexVbo);
		GL30.glDeleteVertexArrays(vao);
	}

	public void draw() {
		if (draw) {
			GL30.glBindVertexArray(vao);
			GL11.glDrawArrays(GL11.GL_TRIANGLES, 0, elements);
		}		
	}
}
