/*-
 * Copyright 2021 Hochschule für Technik Stuttgart
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.hft.stuttgart.citygml.viewer;

import java.nio.FloatBuffer;
import java.nio.IntBuffer;

import org.lwjgl.glfw.GLFW;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL15;
import org.lwjgl.opengl.GL20;
import org.lwjgl.opengl.GL30;
import org.lwjgl.system.MemoryStack;
import org.lwjgl.system.MemoryUtil;

public class ProgressBar {

	private int currentProgress;

	private FloatBuffer clearColor;
	private FloatBuffer clearDepth;

	private int width;
	private int height;

	private long windowId;

	private FloatBuffer box;
	private int vao;
	private int whiteVertexBuffer;
	private int indexBufferName;
	private int colorBufferName;

	public ProgressBar(long windowId, int width, int height) {
		this.windowId = windowId;
		this.width = width;
		this.height = height;
		clearColor = MemoryUtil.memAllocFloat(4);
		clearColor.put(0, 0f).put(1, 0f).put(2, 0f).put(3, 1f);
		clearDepth = MemoryUtil.memAllocFloat(1);
		clearDepth.put(0, 1f);
		box = MemoryUtil.memAllocFloat(8 * 3);
	
		vao = GL30.glGenVertexArrays();
		GL30.glBindVertexArray(vao);
		
		whiteVertexBuffer = GL15.glGenBuffers();
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, whiteVertexBuffer);
		GL20.glVertexAttribPointer(0, 3, GL11.GL_FLOAT, false, 0, 0);
		GL20.glEnableVertexAttribArray(0);
		
		indexBufferName = GL15.glGenBuffers();
		GL15.glBindBuffer(GL15.GL_ELEMENT_ARRAY_BUFFER, indexBufferName);
		try (MemoryStack stack = MemoryStack.stackPush()) {
			// index 
			IntBuffer indexBuffer = stack.mallocInt(12);
			indexBuffer.put(0, 0);
			indexBuffer.put(1, 1);
			indexBuffer.put(2, 2);
			indexBuffer.put(3, 0);
			indexBuffer.put(4, 2);
			indexBuffer.put(5, 3);
			indexBuffer.put(6, 4);
			indexBuffer.put(7, 5);
			indexBuffer.put(8, 6);
			indexBuffer.put(9, 4);
			indexBuffer.put(10, 6);
			indexBuffer.put(11, 7);
			GL15.glBufferData(GL15.GL_ELEMENT_ARRAY_BUFFER, indexBuffer, GL15.GL_STATIC_DRAW);
			
			// color
			FloatBuffer color = stack.mallocFloat(8 * 3);
			for (int i = 0; i < 4 * 3; i++) {
				color.put(i, 1);
			}
			color.put(12, 0);
			color.put(13, 0);
			color.put(14, 1);
			color.put(15, 0);
			color.put(16, 0);
			color.put(17, 1);
			color.put(18, 0);
			color.put(19, 0);
			color.put(20, 1);
			color.put(21, 0);
			color.put(22, 0);
			color.put(23, 1);
			GL20.glEnableVertexAttribArray(1);
			colorBufferName = GL15.glGenBuffers();
			GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, colorBufferName);
			GL15.glBufferData(GL15.GL_ARRAY_BUFFER, color, GL15.GL_STATIC_DRAW);
			GL20.glVertexAttribPointer(1, 3, GL11.GL_FLOAT, false, 0, 0);
		}
		GL30.glBindVertexArray(0);
		
	}

	public void updateWindowSize(int width, int height) {
		this.width = width;
		this.height = height;
		forceDraw(currentProgress);
	}

	private void draw() {
	}

	public void drawProgress(int progress) {
		if (progress == currentProgress) {
			return;
		}
		forceDraw(progress);
	}

	private void forceDraw(int progress) {
		GL30.glClearBufferfv(GL11.GL_COLOR, 0, clearColor);
		GL30.glClearBufferfv(GL11.GL_DEPTH, 0, clearDepth);

		float progressX = width / 2f - 150f;
		float progressY = height / 2f - 30f;
		float progressWidth = 300;
		float progressHeight = 60;
		float okWidth = (progress / 100f) * progressWidth;

		box.put(0, progressX + okWidth);
		box.put(1, progressY);
		box.put(2, 0);
		box.put(3, progressX + progressWidth);
		box.put(4, progressY);
		box.put(5, 0);

		box.put(6, progressX + progressWidth);
		box.put(7, progressY + progressHeight);
		box.put(8, 0);
		box.put(9, progressX + okWidth);
		box.put(10, progressY + progressHeight);
		box.put(11, 0);
		
		box.put(12, progressX);
		box.put(13, progressY);
		box.put(14, 0);
		box.put(15, progressX + okWidth);
		box.put(16, progressY);
		box.put(17, 0);
		box.put(18, progressX + okWidth);
		box.put(19, progressY + progressHeight);
		box.put(20, 0);
		box.put(21, progressX);
		box.put(22, progressY + progressHeight);
		box.put(23, 0);
		
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, whiteVertexBuffer);
		GL15.glBufferData(GL15.GL_ARRAY_BUFFER, box, GL15.GL_STATIC_DRAW);

		GL30.glBindVertexArray(vao);
		GL11.glDrawElements(GL11.GL_TRIANGLES, 12, GL11.GL_UNSIGNED_INT, 0);
		GLFW.glfwSwapBuffers(windowId);
		GLFW.glfwPollEvents();
		currentProgress = progress;

		draw();
	}

	public void free() {
		MemoryUtil.memFree(clearDepth);
		MemoryUtil.memFree(clearColor);
		MemoryUtil.memFree(box);
		GL15.glDeleteBuffers(whiteVertexBuffer);
		GL15.glDeleteBuffers(indexBufferName);
		GL15.glDeleteBuffers(colorBufferName);
		GL30.glDeleteVertexArrays(vao);
	}

}
