/*-
 * Copyright 2021 Hochschule für Technik Stuttgart
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.hft.stuttgart.citygml.viewer.datastructure;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import com.jogamp.opengl.GL;
import com.jogamp.opengl.glu.GLU;
import com.jogamp.opengl.glu.GLUtessellator;
import com.jogamp.opengl.glu.GLUtessellatorCallbackAdapter;

import de.hft.stuttgart.citygml.viewer.math.Vector3d;

/**
 * Tesselator to create triangles out of polygons.
 * 
 * @author Matthias Betz
 *
 */
public class JoglTesselator {

	private static final String DATA_WAS_NOT_A_LIST = "data was not a list";
	private static final GLU glu = new GLU();
	private static final GLUtessellator tess = GLU.gluNewTess();
	private static final GLUtessellatorCallbackAdapter callback;
	private static final ArrayList<Primitive> primitives = new ArrayList<>();

	static {
		callback = new GLUtessellatorCallbackAdapter() {

			private Primitive lastPrimitive;

			@Override
			public void beginData(int type, Object data) {
				if (!(data instanceof ArrayList<?>)) {
					throw new IllegalStateException(DATA_WAS_NOT_A_LIST);
				}
				@SuppressWarnings("unchecked")
				ArrayList<Vector3d> vertices = (ArrayList<Vector3d>) data;
				lastPrimitive = new Primitive(type, vertices);
				primitives.add(lastPrimitive);
			}

			@Override
			public void vertexData(Object vertex, Object data) {
				if (!(data instanceof ArrayList)) {
					throw new IllegalStateException(DATA_WAS_NOT_A_LIST);
				}
				if (!(vertex instanceof Integer)) {
					throw new IllegalStateException("vertex has to be an Integer");
				}
				lastPrimitive.addIndex((Integer) vertex);
			}

			@Override
			public void combineData(double[] coords, Object[] vertexIndices, float[] weights, Object[] outgoingData,
					Object data) {
				if (!(data instanceof ArrayList<?>)) {
					throw new IllegalStateException(DATA_WAS_NOT_A_LIST);
				}
				@SuppressWarnings("unchecked")
				ArrayList<Vector3d> vertices = (ArrayList<Vector3d>) data;
				// to ensure the coordinates are valid
				double[] myCoords = Arrays.copyOf(coords, 3);
				Vector3d newVertex = new Vector3d(myCoords);
				outgoingData[0] = vertices.size();
				vertices.add(newVertex);
			}

			@Override
			public void errorData(int errnum, Object data) {
				throw new TesselationException("GLU Error " + glu.gluErrorString(errnum));
			}

		};
		GLU.gluTessCallback(tess, GLU.GLU_TESS_BEGIN_DATA, callback);
		GLU.gluTessCallback(tess, GLU.GLU_TESS_VERTEX_DATA, callback);
		GLU.gluTessCallback(tess, GLU.GLU_TESS_ERROR_DATA, callback);
		GLU.gluTessCallback(tess, GLU.GLU_TESS_COMBINE_DATA, callback);
		GLU.gluTessCallback(tess, GLU.GLU_TESS_END_DATA, callback);
	}

	public static TesselatedPolygon tesselatePolygon(Polygon p) {
		ArrayList<Vector3d> vertices = new ArrayList<>();
		Vector3d normal = p.calculateNormal();
		if (p.getExteriorRing().getVertices().isEmpty()) {
			return new TesselatedPolygon(vertices, Collections.emptyList());
		}
		synchronized (tess) {
			GLU.gluTessProperty(tess, GLU.GLU_TESS_BOUNDARY_ONLY, GL.GL_FALSE);
			GLU.gluTessBeginPolygon(tess, vertices);
			GLU.gluTessNormal(tess, normal.getX(), normal.getY(), normal.getZ());

			// exterior ring
			Ring ext = p.getExteriorRing();
			createContour(vertices, ext);

			// interior rings
			for (Ring lr : p.getInteriorRings()) {
				createContour(vertices, lr);
			}

			GLU.gluTessEndPolygon(tess);
			ArrayList<Integer> indices = new ArrayList<>();
			for (Primitive primitive : primitives) {
				primitive.fillListWithIndices(indices);
			}
			primitives.clear();
			return new TesselatedPolygon(vertices, indices);
		}
	}

	private static void createContour(List<Vector3d> vertices, Ring lr) {
		GLU.gluTessBeginContour(tess);
		for (Vector3d v : lr.getVertices()) {
			vertices.add(v);
			GLU.gluTessVertex(tess, v.getCoordinates(), 0, vertices.size() - 1);
		}
		GLU.gluTessEndContour(tess);
	}

	private JoglTesselator() {
		// only static usage
	}

}
