/*-
 * Copyright 2021 Hochschule für Technik Stuttgart
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.hft.stuttgart.citygml.viewer.math;

import java.util.ArrayList;
import java.util.List;

import de.hft.stuttgart.citygml.viewer.datastructure.BoundingBox;
import de.hft.stuttgart.citygml.viewer.datastructure.Polygon;
import de.hft.stuttgart.citygml.viewer.parser.FeatureMapper;

/**
 * Utility class for calculating axis aligned bounding boxes for different
 * inputs.
 * 
 * @author Matthias Betz
 *
 */
public class BoundingBoxCalculator {

	private BoundingBoxCalculator() {
		// only static use
	}

	/**
	 * This will calculate two points where every other point in the geometry is
	 * between those two. If the geometry does not contain points or is null, an
	 * empty array is returned. This is considered an axis aligned bounding box.
	 * Only the exterior rings of the polygons is used as inner rings should be
	 * within the exterior or they are faulty.
	 * 
	 * @param a list of polygons containing the vertices
	 * @return an BoundingBox containing the two end points of the bounding box.
	 */
	public static BoundingBox calculateBoundingBox(List<Polygon> polygons) {
		double lowX = Double.MAX_VALUE;
		double highX = Double.NEGATIVE_INFINITY;
		double lowY = Double.MAX_VALUE;
		double highY = Double.NEGATIVE_INFINITY;
		double lowZ = Double.MAX_VALUE;
		double highZ = Double.NEGATIVE_INFINITY;
		for (Polygon p : polygons) {
			// only need to check exterior rings
			for (Vector3d v : p.getExteriorRing().getVertices()) {
				if (v.getX() < lowX) {
					lowX = v.getX();
				}
				if (v.getX() > highX) {
					highX = v.getX();
				}
				if (v.getY() < lowY) {
					lowY = v.getY();
				}
				if (v.getY() > highY) {
					highY = v.getY();
				}
				if (v.getZ() < lowZ) {
					lowZ = v.getZ();
				}
				if (v.getZ() > highZ) {
					highZ = v.getZ();
				}
			}
		}
		Vector3d[] result = new Vector3d[2];
		result[0] = new Vector3d(lowX, lowY, lowZ);
		result[1] = new Vector3d(highX, highY, highZ);
		return BoundingBox.of(result);
	}

	public static BoundingBox calculateBoundingBoxFromPoints(List<? extends Vector3d> points) {
		double lowX = Double.MAX_VALUE;
		double highX = Double.NEGATIVE_INFINITY;
		double lowY = Double.MAX_VALUE;
		double highY = Double.NEGATIVE_INFINITY;
		double lowZ = Double.MAX_VALUE;
		double highZ = Double.NEGATIVE_INFINITY;
		// only need to check exterior rings
		for (Vector3d v : points) {
			if (v.getX() < lowX) {
				lowX = v.getX();
			}
			if (v.getX() > highX) {
				highX = v.getX();
			}
			if (v.getY() < lowY) {
				lowY = v.getY();
			}
			if (v.getY() > highY) {
				highY = v.getY();
			}
			if (v.getZ() < lowZ) {
				lowZ = v.getZ();
			}
			if (v.getZ() > highZ) {
				highZ = v.getZ();
			}
		}
		Vector3d[] result = new Vector3d[2];
		result[0] = new Vector3d(lowX, lowY, lowZ);
		result[1] = new Vector3d(highX, highY, highZ);
		return BoundingBox.of(result);
	}

	public static BoundingBox calculateBoundingBox(FeatureMapper mapper) {
		return calculateBoundingBox(mapper.getLod1Polygons(), mapper.getLod2Polygons(), mapper.getLod3Polygons(), mapper.getLod4Polygons());
	}

	public static BoundingBox calculateBoundingBox(List<Polygon> lod1Polygons, List<Polygon> lod2Polygons,
			List<Polygon> lod3Polygons, List<Polygon> lod4Polygons) {
		List<List<Polygon>> polygonList = new ArrayList<>();
		polygonList.add(lod1Polygons);
		polygonList.add(lod2Polygons);
		polygonList.add(lod3Polygons);
		polygonList.add(lod4Polygons);

		double lowX = Double.MAX_VALUE;
		double highX = Double.NEGATIVE_INFINITY;
		double lowY = Double.MAX_VALUE;
		double highY = Double.NEGATIVE_INFINITY;
		double lowZ = Double.MAX_VALUE;
		double highZ = Double.NEGATIVE_INFINITY;
		for (List<Polygon> polygons : polygonList) {

			for (Polygon p : polygons) {
				// only need to check exterior rings
				for (Vector3d v : p.getExteriorRing().getVertices()) {
					if (v.getX() < lowX) {
						lowX = v.getX();
					}
					if (v.getX() > highX) {
						highX = v.getX();
					}
					if (v.getY() < lowY) {
						lowY = v.getY();
					}
					if (v.getY() > highY) {
						highY = v.getY();
					}
					if (v.getZ() < lowZ) {
						lowZ = v.getZ();
					}
					if (v.getZ() > highZ) {
						highZ = v.getZ();
					}
				}
			}
		}
		Vector3d[] result = new Vector3d[2];
		result[0] = new Vector3d(lowX, lowY, lowZ);
		result[1] = new Vector3d(highX, highY, highZ);
		return BoundingBox.of(result);
	}
}
