/*-
 * Copyright 2021 Hochschule für Technik Stuttgart
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.hft.stuttgart.citygml.viewer.parser;

import java.io.Serializable;

import org.locationtech.proj4j.BasicCoordinateTransform;
import org.locationtech.proj4j.CRSFactory;
import org.locationtech.proj4j.CoordinateReferenceSystem;

/**
 * Container class to store the configuration needed to parse a file. Also
 * contains information on transformation of vertices.
 * 
 * @author Matthias Betz
 *
 */
public class ParserConfiguration implements Serializable {

	private static final long serialVersionUID = 6209047092991074661L;

	private static final CRSFactory CRS_FACTORY = new CRSFactory();

	private transient BasicCoordinateTransform targetTransform = null;
	private transient BasicCoordinateTransform originalTransform = null;

	private String targetTransformString;
	private String originalTransformString;
	private boolean hasTransformation = false;

	public void setCoordinateSystem(CoordinateReferenceSystem crs, CoordinateReferenceSystem tgtCrs) {
		if (crs != null && tgtCrs != null) {
			hasTransformation = true;
			targetTransformString = tgtCrs.getParameterString();
			originalTransformString = crs.getParameterString();
			targetTransform = new BasicCoordinateTransform(crs, tgtCrs);
			originalTransform = new BasicCoordinateTransform(tgtCrs, crs);
		}
	}

	public BasicCoordinateTransform getTargetTransform() {
		if (hasTransformation && targetTransform == null) {
			createCoordinateTransforms();
		}
		return targetTransform;
	}

	private void createCoordinateTransforms() {
		CoordinateReferenceSystem tgtCrs = null;
		CoordinateReferenceSystem crs = null;
		synchronized (CRS_FACTORY) {
			tgtCrs = CRS_FACTORY.createFromParameters("Target", targetTransformString);
			crs = CRS_FACTORY.createFromParameters("Original", originalTransformString);
		}
		targetTransform = new BasicCoordinateTransform(crs, tgtCrs);
		originalTransform = new BasicCoordinateTransform(tgtCrs, crs);
	}

	public BasicCoordinateTransform getOriginalTransform() {
		if (hasTransformation && originalTransform == null) {
			createCoordinateTransforms();
		}
		return originalTransform;
	}
}
